from .iterator import Span, RawIterator


class Token:
    def __init__(self, start, end):
        self.start = start.copy()
        self.end = end.copy()

    @property
    def raw(self):
        return str(Span(RawIterator(self.start), RawIterator(self.end)))

    def __str__(self):
        return str(Span(self.start, self.end))


class EmptyToken(Token):
    def __init__(self):
        super().__init__(RawIterator(''), RawIterator(''))

        print(self)


class StartToken(Token):
    pass


class EndToken(Token):
    pass



from typing import List

import pytest

from quom.tokenizer import CommentToken, CppCommentToken, CCommentToken, NumberToken, PreprocessorToken, \
    PreprocessorIncludeToken, PreprocessorUnknownIncludeToken, QuoteToken, SingleQuoteToken, DoubleQuoteToken, \
    RemainingToken, WhitespaceToken, WhitespaceWhitespaceToken, LinebreakWhitespaceToken, PreprocessorPragmaToken, \
    PreprocessorPragmaOnceToken, PreprocessorDefineToken, PreprocessorIfNotDefinedToken, PreprocessorEndIfToken, \
    tokenize, TokenizeError, Token, StartToken, EndToken


def check_tokens(tokens: List[Token], res):
    res = [StartToken] + res + [EndToken]

    for token, token_type in zip(tokens, res):
        assert isinstance(token, token_type)


def test_comments_cpp_style():
    tokens = tokenize('//abc')
    check_tokens(tokens, [(CommentToken, CppCommentToken)])
    assert str(tokens[1]) == '//abc'
    assert str(tokens[1].content) == 'abc'

    tokens = tokenize('//a\n')
    check_tokens(tokens, [CppCommentToken, WhitespaceToken])
    assert str(tokens[1]) == '//a'

    tokens = tokenize('//\n')
    check_tokens(tokens, [CppCommentToken, WhitespaceToken])
    assert str(tokens[1]) == '//'

    tokens = tokenize(' //abc\n')
    check_tokens(tokens, [WhitespaceWhitespaceToken, CppCommentToken, LinebreakWhitespaceToken])
    assert str(tokens[2]) == '//abc'

    tokens = tokenize('//abc \\\n \\b')
    check_tokens(tokens, [CppCommentToken])


def test_comments_c_style():
    tokens = tokenize('/*ab*/')
    check_tokens(tokens, [(CommentToken, CCommentToken)])
    assert str(tokens[1]) == '/*ab*/'
    assert str(tokens[1].content) == 'ab'

    tokens = tokenize(' /*ab*/ ')
    check_tokens(tokens, [WhitespaceWhitespaceToken, CCommentToken, WhitespaceWhitespaceToken])
    assert str(tokens[2]) == '/*ab*/'

    tokens = tokenize('/*ab\n*/')
    check_tokens(tokens, [CCommentToken])

    tokens = tokenize('/*ab\\\n*/')
    check_tokens(tokens, [CCommentToken])

    tokens = tokenize('/*ab\\b*/')
    check_tokens(tokens, [CCommentToken])

    with pytest.raises(TokenizeError):
        tokenize('/*a')

    with pytest.raises(TokenizeError):
        tokenize('/*a*')

    with pytest.raises(TokenizeError):
        tokenize('/*a* /')


def test_whitespace():
    tokens = tokenize(' ')
    check_tokens(tokens, [WhitespaceWhitespaceToken])
    check_tokens(tokens, [WhitespaceToken])

    tokens = tokenize(' \t')
    check_tokens(tokens, [WhitespaceWhitespaceToken])

    tokens = tokenize('\t')
    check_tokens(tokens, [WhitespaceWhitespaceToken])

    tokens = tokenize('\v')
    check_tokens(tokens, [WhitespaceWhitespaceToken])

    tokens = tokenize('\f')
    check_tokens(tokens, [WhitespaceWhitespaceToken])

    tokens = tokenize('\r')
    check_tokens(tokens, [(WhitespaceToken, LinebreakWhitespaceToken)])

    tokens = tokenize('\n')
    check_tokens(tokens, [(WhitespaceToken, LinebreakWhitespaceToken)])

    tokens = tokenize('\r\n')
    check_tokens(tokens, [LinebreakWhitespaceToken])

    tokens = tokenize('\r\n\r')
    check_tokens(tokens, [LinebreakWhitespaceToken, LinebreakWhitespaceToken])

    tokens = tokenize('\r\n\n')
    check_tokens(tokens, [LinebreakWhitespaceToken, LinebreakWhitespaceToken])

    tokens = tokenize('\r\n\r\n\n')
    check_tokens(tokens, [LinebreakWhitespaceToken, LinebreakWhitespaceToken, LinebreakWhitespaceToken])

    tokens = tokenize('\n\r\r')
    check_tokens(tokens, [LinebreakWhitespaceToken, LinebreakWhitespaceToken, LinebreakWhitespaceToken])

    tokens = tokenize(' \n ')
    check_tokens(tokens, [WhitespaceWhitespaceToken, LinebreakWhitespaceToken, WhitespaceWhitespaceToken])


def test_quote_single():
    tokens = tokenize("'abc'")
    check_tokens(tokens, [SingleQuoteToken])
    check_tokens(tokens, [QuoteToken])
    assert str(tokens[1]) == "'abc'"

    tokens = tokenize(" 'abc' ")
    check_tokens(tokens, [WhitespaceWhitespaceToken, SingleQuoteToken, WhitespaceWhitespaceToken])
    assert str(tokens[2]) == "'abc'"

    tokens = tokenize("'a\\bc'")
    check_tokens(tokens, [SingleQuoteToken])

    tokens = tokenize("'a\\'bc'")
    check_tokens(tokens, [SingleQuoteToken])

    with pytest.raises(TokenizeError):
        tokenize("'a")

    with pytest.raises(TokenizeError):
        tokenize("'")


def test_quote_double():
    tokens = tokenize('"abc"')
    check_tokens(tokens, [(QuoteToken, DoubleQuoteToken)])
    assert str(tokens[1]) == '"abc"'

    tokens = tokenize(' "abc" ')
    check_tokens(tokens, [WhitespaceWhitespaceToken, DoubleQuoteToken, WhitespaceWhitespaceToken])
    assert str(tokens[2]) == '"abc"'

    tokens = tokenize('"abc\\""')
    check_tokens(tokens, [DoubleQuoteToken])
    assert str(tokens[1]) == '"abc\\""'

    tokens = tokenize('R"(abc)"')
    check_tokens(tokens, [RemainingToken, DoubleQuoteToken])
    assert str(tokens[2]) == '"(abc)"'

    tokens = tokenize('R"(a\b\\r\\abc)"')
    check_tokens(tokens, [RemainingToken, DoubleQuoteToken])
    assert str(tokens[2]) == '"(a\b\\r\\abc)"'

    tokens = tokenize('R"1=#(abc)1=#abc)1=#"')
    check_tokens(tokens, [RemainingToken, DoubleQuoteToken])
    assert tokens[2].is_raw_encoding

    tokens = tokenize('u8R"(abc)"')
    check_tokens(tokens, [RemainingToken, DoubleQuoteToken])
    assert tokens[2].is_raw_encoding

    tokens = tokenize('uR"(abc)"')
    check_tokens(tokens, [RemainingToken, DoubleQuoteToken])
    assert tokens[2].is_raw_encoding

    tokens = tokenize('UR"(a)bc)"')
    check_tokens(tokens, [RemainingToken, DoubleQuoteToken])
    assert tokens[2].is_raw_encoding

    tokens = tokenize('LR"=(a)bc)="')
    check_tokens(tokens, [RemainingToken, DoubleQuoteToken])
    assert tokens[2].is_raw_encoding

    tokens = tokenize('u"abc"')
    check_tokens(tokens, [RemainingToken, DoubleQuoteToken])

    tokens = tokenize('u8"abc"')
    check_tokens(tokens, [RemainingToken, DoubleQuoteToken])

    tokens = tokenize('U"abc"')
    check_tokens(tokens, [RemainingToken, DoubleQuoteToken])

    tokens = tokenize('L"abc"')
    check_tokens(tokens, [RemainingToken, DoubleQuoteToken])

    with pytest.raises(TokenizeError):
        tokenize('"a')

    with pytest.raises(TokenizeError):
        tokenize('a"a')

    with pytest.raises(TokenizeError):
        tokenize('R"a"')

    with pytest.raises(TokenizeError):
        tokenize('R"(a"')


def test_number():
    tokens = tokenize('0')
    check_tokens(tokens, [NumberToken])
    assert str(tokens[1]) == '0'

    tokens = tokenize('123')
    check_tokens(tokens, [NumberToken])
    assert str(tokens[1]) == '123'

    tokens = tokenize("1'23")
    check_tokens(tokens, [NumberToken])
    assert str(tokens[1]) == "1'23"

    tokens = tokenize(' 1234567890 ')
    check_tokens(tokens, [WhitespaceWhitespaceToken, NumberToken, WhitespaceWhitespaceToken])
    assert str(tokens[2]) == '1234567890'

    tokens = tokenize('001.1')
    check_tokens(tokens, [NumberToken])

    tokens = tokenize('001e1')
    check_tokens(tokens, [NumberToken])

    tokens = tokenize('0x')
    check_tokens(tokens, [NumberToken])

    tokens = tokenize('1xx13b')
    check_tokens(tokens, [NumberToken])

    tokens = tokenize('.1')
    check_tokens(tokens, [NumberToken])
    assert str(tokens[1]) == '.1'

    tokens = tokenize('.123')
    check_tokens(tokens, [NumberToken])

    tokens = tokenize('123.345')
    check_tokens(tokens, [NumberToken])

    tokens = tokenize('123.345e3')
    check_tokens(tokens, [NumberToken])

    tokens = tokenize('123e+4')
    check_tokens(tokens, [NumberToken])

    tokens = tokenize('123e-4')
    check_tokens(tokens, [NumberToken])

    tokens = tokenize("12'3.345")
    check_tokens(tokens, [NumberToken])

    tokens = tokenize('01e1')
    check_tokens(tokens, [NumberToken])

    tokens = tokenize('01.1')
    check_tokens(tokens, [NumberToken])

    with pytest.raises(TokenizeError):
        tokenize("12'")

    tokens = tokenize('0x1')
    check_tokens(tokens, [NumberToken])

    tokens = tokenize('0xFp1')
    check_tokens(tokens, [NumberToken])

    tokens = tokenize("0x02'3")
    check_tokens(tokens, [NumberToken])

    tokens = tokenize('0x023p+1')
    check_tokens(tokens, [NumberToken])

    tokens = tokenize('0xABCDEFabcdef')
    check_tokens(tokens, [NumberToken])

    tokens = tokenize('0x0.123p-1')
    check_tokens(tokens, [NumberToken])

    tokens = tokenize("0x0.e2'3p-1'0")
    check_tokens(tokens, [NumberToken])

    tokens = tokenize('0x0.p1')
    check_tokens(tokens, [NumberToken])

    tokens = tokenize('0xa.Ap1')
    check_tokens(tokens, [NumberToken])

    tokens = tokenize('0xA.ap1')
    check_tokens(tokens, [NumberToken])

    with pytest.raises(TokenizeError):
        tokenize("0x12'")

    tokens = tokenize('0x0.123')
    check_tokens(tokens, [NumberToken])

    tokens = tokenize('0x0.123p-A')
    check_tokens(tokens, [NumberToken])

    tokens = tokenize('0b1')
    check_tokens(tokens, [NumberToken])

    tokens = tokenize('0b01')
    check_tokens(tokens, [NumberToken])

    tokens = tokenize('0b010101')
    check_tokens(tokens, [NumberToken])

    with pytest.raises(TokenizeError):
        tokenize("0b01'")

    tokens = tokenize('01')
    check_tokens(tokens, [NumberToken])

    tokens = tokenize("0'1")
    check_tokens(tokens, [NumberToken])

    tokens = tokenize("0012345'67")
    check_tokens(tokens, [NumberToken])

    tokens = tokenize('01x')
    check_tokens(tokens, [NumberToken])

    tokens = tokenize("0012345'68")
    check_tokens(tokens, [NumberToken])

    tokens = tokenize("12'd.z'.x1.'1")
    check_tokens(tokens, [NumberToken])


def test_preprocessor():
    tokens = tokenize('#')
    check_tokens(tokens, [PreprocessorToken])

    tokens = tokenize('#\n ')
    check_tokens(tokens, [PreprocessorToken, WhitespaceWhitespaceToken])

    tokens = tokenize('# /**/\n')
    check_tokens(tokens, [PreprocessorToken])
    check_tokens(tokens[1].preprocessor_tokens,
                 [RemainingToken, WhitespaceWhitespaceToken, CCommentToken, LinebreakWhitespaceToken])

    tokens = tokenize('#pragma')
    check_tokens(tokens, [PreprocessorToken])

    tokens = tokenize(' #pragma')
    check_tokens(tokens, [WhitespaceWhitespaceToken, PreprocessorToken])

    tokens = tokenize('#pragma once')
    check_tokens(tokens, [(PreprocessorPragmaToken, PreprocessorPragmaOnceToken)])

    tokens = tokenize('#pragma /*abc*/ once /*def*/')
    check_tokens(tokens, [PreprocessorPragmaToken])

    tokens = tokenize('#pragma once abc')
    check_tokens(tokens, [PreprocessorPragmaToken])

    tokens = tokenize('#include "abc" ')
    check_tokens(tokens, [(PreprocessorToken, PreprocessorIncludeToken)])
    assert str(tokens[1].path) == 'abc'

    tokens = tokenize('#include "abc\\"" ')
    check_tokens(tokens, [PreprocessorToken])
    assert str(tokens[1].path) == 'abc\\"'

    tokens = tokenize('#include /*123*/ <abc> ')
    check_tokens(tokens, [PreprocessorToken])
    assert str(tokens[1].path) == 'abc'

    with pytest.raises(TokenizeError):
        tokenize('#include "abc\\" ')

    with pytest.raises(TokenizeError):
        tokenize('#include <abc')

    tokens = tokenize('#include ')
    check_tokens(tokens, [PreprocessorUnknownIncludeToken])

    tokens = tokenize('#define')
    check_tokens(tokens, [PreprocessorDefineToken])

    tokens = tokenize('#define a(x) auto a##x = #x')
    check_tokens(tokens, [PreprocessorDefineToken])
    check_tokens(tokens[1].preprocessor_tokens,
                 [RemainingToken, RemainingToken, WhitespaceWhitespaceToken, RemainingToken, WhitespaceWhitespaceToken,
                  RemainingToken, WhitespaceWhitespaceToken, RemainingToken, WhitespaceWhitespaceToken, RemainingToken,
                  WhitespaceWhitespaceToken, RemainingToken])

    res = ''
    for token in tokens[1].preprocessor_tokens:
        res += str(token)
    assert res == '#define a(x) auto a##x = #x'

    tokens = tokenize('#define eprintf(format, ...) fprintf (stderr, format, __VA_ARGS__)')
    check_tokens(tokens, [PreprocessorDefineToken])

    tokens = tokenize('# /* abc */ define  /* test */')
    check_tokens(tokens, [PreprocessorDefineToken])

    tokens = tokenize('#ifndef asd')
    check_tokens(tokens, [PreprocessorIfNotDefinedToken])

    tokens = tokenize('#endif /*asd*/')
    check_tokens(tokens, [PreprocessorEndIfToken])
    assert len(tokens[1].preprocessor_instruction) == 2
    assert isinstance(tokens[1].preprocessor_instruction[0], RemainingToken)
    assert isinstance(tokens[1].preprocessor_instruction[1], RemainingToken)
    assert len(tokens[1].preprocessor_instruction) == 2
    assert isinstance(tokens[1].preprocessor_arguments[0], WhitespaceWhitespaceToken)
    assert isinstance(tokens[1].preprocessor_arguments[1], CCommentToken)


def test_remaining():
    tokens = tokenize('abc')
    check_tokens(tokens, [RemainingToken])
    assert str(tokens[1]) == 'abc'

    tokens = tokenize('abc ')
    check_tokens(tokens, [RemainingToken, WhitespaceWhitespaceToken])
    assert str(tokens[1]) == 'abc'

    tokens = tokenize(' abc ')
    check_tokens(tokens, [WhitespaceWhitespaceToken, RemainingToken, WhitespaceWhitespaceToken])
    assert str(tokens[2]) == 'abc'

    tokens = tokenize('_ab_c')
    check_tokens(tokens, [RemainingToken])
    assert str(tokens[1]) == '_ab_c'

    for symbol in '+-*/%<>&!=?.,[]{}():|;~^':
        tokens = tokenize(symbol)
        check_tokens(tokens, [RemainingToken])

    tokens = tokenize('kwqjj8a8gja98gj9\b\1\123')
    check_tokens(tokens, [RemainingToken])



from typing import List

from .iterator import RawIterator, LineWrapIterator
from .remaining_tokenizer import RemainingToken
from .token import Token
from .tokenize_error import TokenizeError


class QuoteToken(Token):
    pass


class SingleQuoteToken(QuoteToken):
    pass


class DoubleQuoteToken(QuoteToken):
    def __init__(self, start, end, is_raw_encdoing):
        super().__init__(start, end)
        self.is_raw_encoding = is_raw_encdoing


def check_is_raw_encoding(token: Token):
    if isinstance(token, RemainingToken):
        return str(token) in ['R', 'u8R', 'LR', 'uR', 'UR']
    return False


def scan_for_quote_single(tokens: List[Token], it: LineWrapIterator):
    if it.curr != "'":
        return False
    start = it.copy()

    # Parse until non escaped '.
    backslashes = 0
    while it.next() and (it.curr != "'" or backslashes % 2 != 0):
        if it.curr == '\\':
            backslashes += 1
        else:
            backslashes = 0

    # Check if end of file is reached.
    if it.curr != "'":
        raise TokenizeError('Character sequence not terminated!', it)
    it.next()

    tokens.append(SingleQuoteToken(start, it))
    return True


def scan_for_quote_double(tokens: List[Token], it: LineWrapIterator):
    if it.curr != '"':
        return None
    start = it.copy()

    is_raw_encoding = check_is_raw_encoding(tokens[-1])
    if not is_raw_encoding:
        # Parse until end of line or non escaped ".
        backslashes = 0
        while it.next() and (it.curr != '"' or backslashes % 2 != 0):
            if it.curr == '\\':
                backslashes += 1
            else:
                backslashes = 0

        # Check if end of file is reached.
        if it.curr != '"':
            raise TokenizeError('Character sequence not terminated!', it)
        it.next()
    else:
        delimiter = ''

        # Parse until end of introductory delimiter.
        while it.next() and it.curr != '(' and it.curr:
            delimiter += it.curr

        if it.curr != '(':
            raise TokenizeError('No introductory delimiter inside raw string literal found!', it)

        # Define terminating delimiter.
        delimiter = ')' + delimiter + '"'
        it = RawIterator(it)

        # Parse until delimiter occours again.
        # TODO: Please optimize me.
        string = ''
        while it.next():
            string += it.curr
            if len(string) > len(delimiter) and string.endswith(delimiter):
                break

        if it.curr != '"':
            raise TokenizeError('No terminating delimiter inside raw string literal found!', it)
        it.next()

    tokens.append(DoubleQuoteToken(start, it, is_raw_encoding))
    return True


def scan_for_quote(tokens: List[Token], it: LineWrapIterator):
    if not scan_for_quote_double(tokens, it):
        return scan_for_quote_single(tokens, it)
    return True



import os
from io import StringIO
from pathlib import Path

from quom import Quom
from quom.__main__ import main

FILE_MAIN_HPP = """
int foo = 3;

int foo();
"""

FILE_MAIN_CPP = """
int foo() { return 42; }
"""

RESULT = """
int foo = 3;

int foo();

int foo() { return 42; }
"""


def test_source_directory(fs):
    os.makedirs('project/')
    os.chdir('project/')
    os.makedirs('include/')
    os.makedirs('src/')

    with open('include/main.hpp', 'w+') as file:
        file.write(FILE_MAIN_HPP)

    with open('src/main.cpp', 'w+') as file:
        file.write(FILE_MAIN_CPP)

    dst = StringIO()
    Quom(Path('include/main.hpp'), dst)
    assert dst.getvalue() != RESULT

    dst = StringIO()
    Quom(Path('include/main.hpp'), dst, relative_source_directories=[Path('../src')])
    assert dst.getvalue() == RESULT

    dst = StringIO()
    Quom(Path('include/main.hpp'), dst, source_directories=[Path('src').resolve()])
    assert dst.getvalue() == RESULT

    dst = StringIO()
    Quom(Path('include/main.hpp'), dst, source_directories=[Path('/project/src')])
    assert dst.getvalue() == RESULT

    main(['include/main.hpp', 'result.hpp', '-S', './../src'])
    assert Path('result.hpp').read_text() == RESULT

    main(['include/main.hpp', 'result.hpp', '-S', 'src'])
    assert Path('result.hpp').read_text() == RESULT

    main(['include/main.hpp', 'result.hpp', '-S', '/project/src'])
    assert Path('result.hpp').read_text() == RESULT



import os
from io import StringIO
from pathlib import Path

import pytest

from quom import Quom, QuomError

FILE_MAIN_HPP = """
int foo = 3;

#include "my_lib/core/core.hpp"
#include "my_other_lib/bar.hpp"
"""

FILE_CORE_HPP = """
#pragma once

#include "my_lib/util/foo.hpp"
"""

FILE_CORE_CPP = """
#include "my_lib/core/core.hpp"

int i = 10;
"""

FILE_FOO_HPP = """
#pragma once

int foo();
"""

FILE_FOO_CPP = """
#include "foo.hpp"

int foo() { return 42; }
"""

FILE_BAR_HPP = """
#pragma once

float bar();
"""

FILE_INFO_HPP = """

float f = 0.42;

"""

FILE_BAR_CPP = """
#include "bar.hpp"

#include "info.hpp"

float bar() { return f; }
"""

RESULT = """
int foo = 3;

int foo();

float bar();

int foo() { return 42; }

int i = 10;

float f = 0.42;

float bar() { return f; }
"""


def test_include_directory(fs):
    os.makedirs('include/my_lib/core')
    os.makedirs('include/my_lib/util')
    os.makedirs('include/my_other_lib/')

    with open('include/my_lib/main.hpp', 'w+') as file:
        file.write(FILE_MAIN_HPP)

    with open('include/my_lib/core/core.hpp', 'w+') as file:
        file.write(FILE_CORE_HPP)

    with open('include/my_lib/core/core.cpp', 'w+') as file:
        file.write(FILE_CORE_CPP)

    with open('include/my_lib/util/foo.hpp', 'w+') as file:
        file.write(FILE_FOO_HPP)

    with open('include/my_lib/util/foo.cpp', 'w+') as file:
        file.write(FILE_FOO_CPP)

    with open('include/my_other_lib/bar.hpp', 'w+') as file:
        file.write(FILE_BAR_HPP)

    with open('include/my_other_lib/bar.cpp', 'w+') as file:
        file.write(FILE_BAR_CPP)

    with open('include/my_other_lib/info.hpp', 'w+') as file:
        file.write(FILE_INFO_HPP)

    dst = StringIO()

    with pytest.raises(QuomError):
        Quom(Path('include/my_lib/main.hpp'), dst)

    dst = StringIO()

    Quom(Path('include/my_lib/main.hpp'), dst, include_directories=['include/'])

    assert dst.getvalue() == RESULT



from io import StringIO
from pathlib import Path

import pytest

from quom import Quom, QuomError
from quom.__main__ import main

FILE_MAIN_HPP = """\
#pragma once

#ifndef FOOBAR_HPP
#define FOOBAR_HPP

#include "foo.hpp"

#endif // FOOBAR_HPP
"""

FILE_FOO_HPP = """\
#pragma once

# /* */ ifndef /*123*/ FOOBAR_FOO_HPP
#define FOOBAR_FOO_HPP // abc

#include <iostream>

extern int foo;

#endif // FOOBAR_FOO_HPP
"""

FILE_FOO_CPP = """\
#include "foo.hpp"

#include <algorithm>

int foo = 42;
"""

RESULT_NORMAL = """\
#pragma once

#ifndef FOOBAR_HPP
#define FOOBAR_HPP

# /* */ ifndef /*123*/ FOOBAR_FOO_HPP
#define FOOBAR_FOO_HPP // abc

#include <iostream>

extern int foo;

#endif // FOOBAR_FOO_HPP

#endif // FOOBAR_HPP

#include <algorithm>

int foo = 42;
"""

RESULT_NORMAL_WITHOUT_TRIM = """\
#pragma once

#ifndef FOOBAR_HPP
#define FOOBAR_HPP



# /* */ ifndef /*123*/ FOOBAR_FOO_HPP
#define FOOBAR_FOO_HPP // abc

#include <iostream>

extern int foo;

#endif // FOOBAR_FOO_HPP


#endif // FOOBAR_HPP


#include <algorithm>

int foo = 42;
"""

RESULT_NORMAL_WITHOUT_SOURCES = """\
#pragma once

#ifndef FOOBAR_HPP
#define FOOBAR_HPP

# /* */ ifndef /*123*/ FOOBAR_FOO_HPP
#define FOOBAR_FOO_HPP // abc

#include <iostream>

extern int foo;

#endif // FOOBAR_FOO_HPP

#endif // FOOBAR_HPP
"""

RESULT_WITH_INCLUDE_GUARD_FORMAT = """\
#pragma once

#ifndef FOOBAR_HPP
#define FOOBAR_HPP

#include <iostream>

extern int foo;

#endif // FOOBAR_HPP

#include <algorithm>

int foo = 42;
"""

FILE_MAIN_WITH_STITCH_LOCATION_HPP = """\
#pragma once

#ifndef FOOBAR_HPP
#define FOOBAR_HPP

#include "foo.hpp"

#endif // FOOBAR_HPP

#ifdef MAIN

// ~> stitch <~

#endif
"""

RESULT_WITH_STITCH_LOCATION = """\
#pragma once

#ifndef FOOBAR_HPP
#define FOOBAR_HPP

# /* */ ifndef /*123*/ FOOBAR_FOO_HPP
#define FOOBAR_FOO_HPP // abc

#include <iostream>

extern int foo;

#endif // FOOBAR_FOO_HPP

#endif // FOOBAR_HPP

#ifdef MAIN

#include <algorithm>

int foo = 42;

#endif
"""


def init():
    with open('main.hpp', 'w+') as file:
        file.write(FILE_MAIN_HPP)

    with open('foo.hpp', 'w+') as file:
        file.write(FILE_FOO_HPP)

    with open('foo.cpp', 'w+') as file:
        file.write(FILE_FOO_CPP)


def test_normal(fs):
    init()

    dst = StringIO()
    Quom(Path('main.hpp'), dst)

    assert dst.getvalue() == RESULT_NORMAL


def test_normal_without_trim(fs):
    init()

    dst = StringIO()
    Quom(Path('main.hpp'), dst, trim=False)

    assert dst.getvalue() == RESULT_NORMAL_WITHOUT_TRIM


def test_with_include_guard_format(fs):
    init()

    dst = StringIO()
    Quom(Path('main.hpp'), dst, include_guard_format='FOOBAR_.+_HPP')

    assert dst.getvalue() == RESULT_WITH_INCLUDE_GUARD_FORMAT


def test_with_mismatching_include_guard_format(fs):
    init()

    dst = StringIO()
    Quom(Path('main.hpp'), dst, include_guard_format='FOOBAR_.+_HP')

    assert dst.getvalue() == RESULT_NORMAL


def test_with_stitch_location(fs):
    init()
    with open('main.hpp', 'w') as file:
        file.write(FILE_MAIN_WITH_STITCH_LOCATION_HPP)

    dst = StringIO()
    Quom(Path('main.hpp'), dst, stitch_format='~> stitch <~')

    assert dst.getvalue() == RESULT_WITH_STITCH_LOCATION


def test_with_mismatching_stitch(fs):
    init()

    dst = StringIO()
    with pytest.raises(QuomError):
        Quom(Path('main.hpp'), dst, stitch_format='~ stitch ~')


def test_with_missing_header_file(fs):
    init()
    Path('foo.hpp').unlink()

    dst = StringIO()
    with pytest.raises(QuomError):
        Quom(Path('main.hpp'), dst)


def test_with_missing_source_file(fs):
    init()
    Path('foo.cpp').unlink()

    dst = StringIO()
    Quom(Path('main.hpp'), dst)

    assert dst.getvalue() == RESULT_NORMAL_WITHOUT_SOURCES


def test_main(fs):
    init()

    main(['main.hpp', 'result.hpp'])
    assert Path('result.hpp').read_text() == RESULT_NORMAL

    main(['main.hpp', 'result.hpp', '-S', '.'])
    assert Path('result.hpp').read_text() == RESULT_NORMAL



from typing import List

from .comment_tokenizer import scan_for_comment
from .iterator import LineWrapIterator
from .number_tokenizer import scan_for_number
from .preprocessor_tokenizer import scan_for_preprocessor
from .quote_tokenizer import scan_for_quote
from .remaining_tokenizer import scan_for_remaining
from .token import Token, StartToken, EndToken
from .whitespace_tokenizer import scan_for_whitespace


def tokenize(src) -> List[Token]:
    it = LineWrapIterator(src)

    tokens = [StartToken(it, it)]

    while it.curr != '\0':
        succeeded = scan_for_whitespace(tokens, it)
        if not succeeded:
            succeeded = scan_for_comment(tokens, it)
        if not succeeded:
            succeeded = scan_for_quote(tokens, it)
        if not succeeded:
            succeeded = scan_for_number(tokens, it)
        if not succeeded:
            succeeded = scan_for_preprocessor(tokens, it)
        if not succeeded:
            scan_for_remaining(tokens, it)

    tokens.append(EndToken(it, it))

    return tokens



from quom.tokenizer.iterator import RawIterator, LineWrapIterator, Span


def check_iterator(it, res):
    crr = '\0'

    for c in res:
        prv = crr
        crr = c

        assert it.prev == prv
        assert it.curr == crr

        it.next()

    assert it.next() is False


def test_raw_iterator():
    it = RawIterator('ab')
    check_iterator(it, 'ab')

    it = RawIterator('a\rb')
    check_iterator(it, 'a\rb')

    it = RawIterator('a\r\nb')
    check_iterator(it, 'a\r\nb')

    it = RawIterator('a\\\r\nb')
    check_iterator(it, 'a\\\r\nb')

    it = RawIterator('a\\\nb')
    check_iterator(it, 'a\\\nb')

    it = RawIterator('a\\\rb')
    check_iterator(it, 'a\\\rb')

    it = RawIterator('a\\\r\\\r\nb')
    check_iterator(it, 'a\\\r\\\r\nb')

    it = RawIterator('a\\b')
    check_iterator(it, 'a\\b')

    it = RawIterator('"\\a')
    check_iterator(it, '"\\a')

    it = RawIterator('\\\\')
    check_iterator(it, '\\\\')

    it = RawIterator('\\')
    check_iterator(it, '\\')

    it = RawIterator('\\\n')
    check_iterator(it, '\\\n')

    it = RawIterator('\r')
    check_iterator(it, '\r')

    it = RawIterator('a')
    assert it.lookahead == '\0'

    it = RawIterator('\\')
    assert it.lookahead == '\0'

    it = RawIterator('\\\na')
    assert it.lookahead == '\n'

    it = RawIterator('')
    assert it.lookahead == '\0'


def test_escape_iterator():
    it = LineWrapIterator('ab')
    check_iterator(it, 'ab')

    it = LineWrapIterator('a\rb')
    check_iterator(it, 'a\rb')

    it = LineWrapIterator('a\r\nb')
    check_iterator(it, 'a\r\nb')

    it = LineWrapIterator('a\\\r\nb')
    check_iterator(it, 'ab')

    it = LineWrapIterator('a\\\nb')
    check_iterator(it, 'ab')

    it = LineWrapIterator('a\\\rb')
    check_iterator(it, 'ab')

    it = LineWrapIterator('a\\\\\nb')
    check_iterator(it, 'a\\b')

    it = LineWrapIterator('a\\\r\\\r\nb')
    check_iterator(it, 'ab')

    it = LineWrapIterator('a\\b')
    check_iterator(it, 'a\\b')

    it = LineWrapIterator('"\\a')
    check_iterator(it, '"\\a')

    it = LineWrapIterator('\\\\')
    check_iterator(it, '\\\\')

    it = LineWrapIterator('\\')
    check_iterator(it, '\\')

    it = LineWrapIterator('\\\n')
    check_iterator(it, '\0')

    it = LineWrapIterator('a')
    assert it.lookahead == '\0'

    it = LineWrapIterator('\\')
    assert it.lookahead == '\0'

    it = LineWrapIterator('\\\na')
    assert it.lookahead == '\0'

    it = LineWrapIterator('')
    assert it.lookahead == '\0'

    it = LineWrapIterator('a\\\n')
    assert it.lookahead == '\0'


def test_copy():
    it1 = LineWrapIterator('ab')
    assert it1.curr == 'a'

    it2 = it1.copy()
    it2.next()
    assert it1.curr == 'a'
    assert it2.curr == 'b'

    it1.next()
    assert it1.curr == 'b'
    assert it2.curr == 'b'


def test_iterator_casting():
    it = LineWrapIterator('a\\\r\\b\\\nc')
    assert it.curr == 'a'

    it = LineWrapIterator(it)
    it.next()
    assert it.curr == '\\'
    it.next()
    assert it.curr == 'b'

    it = RawIterator(it)
    it.next()
    assert it.curr == '\\'
    it.next()
    assert it.curr == '\n'
    it.next()
    assert it.curr == 'c'


def test_span():
    it1 = LineWrapIterator('abc')
    assert ''.join(it1) == 'abc'
    assert ''.join(it1) == 'abc'

    it2 = it1.copy()
    it2.next()
    assert ''.join(it2) == 'bc'
    assert ''.join(it2) == 'bc'

    it1 = LineWrapIterator('a ')
    it2 = it1.copy()
    it2.next()

    span = Span(it1, it2)
    assert ''.join(span) == 'a'
    assert ''.join(span) == 'a'



from typing import Union


class Iterator:
    pass


class Span:
    def __init__(self, start: 'RawIterator', end: 'RawIterator' = None):
        self.it = start.copy()
        if end and end._it.curr_pos < len(self.it):
            self._length = end._it.curr_pos
        else:
            self._length = len(self.it)

    def __len__(self):
        return self._length

    def __iter__(self):
        tmp = Span(self.it)
        tmp._length = self._length
        return tmp

    def __next__(self):
        if self.it._it.curr_pos >= len(self):
            raise StopIteration()
        tmp = self.it.curr
        self.it.next()
        return tmp

    def __str__(self):
        return ''.join(self)


class Iterable:
    def __init__(self, src):
        self.src = src
        self.length = len(src)
        self.prev = '\0'
        self.curr = '\0'
        self.curr_pos = -1

    def copy(self):
        tmp = Iterable(self.src)
        tmp.prev = self.prev
        tmp.curr = self.curr
        tmp.curr_pos = self.curr_pos
        return tmp


class RawIterator:
    def __init__(self, it: Union[Iterable, 'RawIterator', str]):
        if isinstance(it, Iterable):
            self._it = it
        elif isinstance(it, RawIterator):
            self._it = it._it
        else:
            self._it = Iterable(it)
        # Init iterator if not already done.
        if self._it.curr_pos == -1:
            self.__step()

    @property
    def prev(self):
        return self._it.prev

    @property
    def curr(self):
        return self._it.curr

    @property
    def lookahead(self):
        if self._it.curr_pos + 1 >= self._it.length:
            return '\0'

        nxt = self._step(self._it.src, self._it.curr_pos + 1)
        if 0 <= nxt < self._it.length:
            return self._it.src[nxt]
        return '\0'

    def copy(self):
        return self.__class__(self._it.copy())

    def __len__(self):
        return self._it.length

    def __iter__(self):
        return Span(self)

    def next(self):
        self.__step()
        return True if self._it.curr != '\0' else False

    def __step(self):
        self._it.prev = self._it.curr

        if self._it.curr_pos + 1 >= self._it.length:
            self._it.curr_pos = self._it.length
            self._it.curr = '\0'
            return

        src = self._it.src
        nxt = self._it.curr_pos + 1

        self._it.curr_pos = self._step(src, nxt)

        if 0 <= self._it.curr_pos < self._it.length:
            self._it.curr = self._it.src[self._it.curr_pos]
        else:
            self._it.curr = '\0'

    def _step(self, src, nxt):
        # Get next character.
        return nxt


class LineWrapIterator(RawIterator):
    def _step(self, src, nxt):
        # Get next character, but:
        # * do line wrapping (backslash followed by \r and/or \n)
        while nxt < self._it.length:
            if src[nxt] == '\\':
                if nxt + 1 >= self._it.length:
                    break
                if src[nxt + 1] == '\n':
                    nxt += 2
                    continue
                elif src[nxt + 1] == '\r':
                    nxt += 2
                    if nxt < self._it.length and src[nxt] == '\n':
                        nxt += 1
                    continue
            break
        return nxt



import re
from pathlib import Path
from queue import Queue
from typing import TextIO, Union, List

from .quom_error import QuomError
from .tokenizer import tokenize, Token, CommentToken, PreprocessorToken, PreprocessorIfNotDefinedToken, \
    PreprocessorDefineToken, PreprocessorEndIfToken, PreprocessorIncludeToken, PreprocessorPragmaOnceToken, \
    RemainingToken, LinebreakWhitespaceToken, EmptyToken, StartToken, EndToken, WhitespaceToken

CONTINUOUS_LINE_BREAK_START = 0
CONTINUOUS_BREAK_REACHED = 3


def find_token(tokens: List[Token], token_type: any):
    for i, token in enumerate(tokens):
        if isinstance(token, token_type):
            return i, token
    return None, None


def contains_only_whitespace_and_comment_tokens(tokens: List[Token]):
    for token in tokens:
        if not isinstance(token, (WhitespaceToken, CommentToken, EndToken)):
            return False
    return True


class Quom:
    def __init__(self, src_file_path: Union[Path, str], dst: TextIO, stitch_format: str = None,
                 include_guard_format: str = None, trim: bool = True,
                 include_directories: List[Union[Path, str]] = None,
                 relative_source_directories: List[Union[Path]] = None,
                 source_directories: List[Union[Path]] = None):
        self.__dst = dst
        self.__stitch_format = stitch_format
        self.__include_guard_format = re.compile('^{}$'.format(include_guard_format)) if include_guard_format else None
        self.__trim = trim
        self.__include_directories = [Path(x) for x in include_directories] if include_directories else []
        self.__relative_source_directories = relative_source_directories if relative_source_directories else [] \
            if source_directories else [Path('.')]
        self.__source_directories = source_directories if source_directories else [Path('.')]

        self.__processed_files = set()
        self.__source_files = Queue()
        self.__cont_lb = CONTINUOUS_LINE_BREAK_START
        self.__prev_token = EmptyToken()

        self.__process_file(Path(), src_file_path, False, True)

        if not self.__source_files.empty():
            if stitch_format is not None:
                raise QuomError('Couldn\'t stitch source files. The stitch location "{}" was not found.'
                                .format(stitch_format))
            while not self.__source_files.empty():
                self.__process_file(Path(), self.__source_files.get(), True)
            # Write last token.
            self.__write_token(self.__prev_token, True)
        elif self.__cont_lb == CONTINUOUS_LINE_BREAK_START or not isinstance(self.__prev_token,
                                                                             LinebreakWhitespaceToken):
            # Write last token, if not a continuous line break.
            self.__write_token(self.__prev_token, True)

    def __process_file(self, relative_path: Path, include_path: Path, is_source_file: bool,
                       is_main_header=False):
        # First check if file exists relative.
        file_path = relative_path / include_path
        if file_path.exists():
            with file_path.open() as file:
                tokens = tokenize(file.read())
        else:
            # Otherwise search in include directories.
            for include_directory in self.__include_directories:
                file_path = include_directory / include_path
                if file_path.exists():
                    with file_path.open() as file:
                        tokens = tokenize(file.read())
                    break
            else:
                raise QuomError('Include not found: "{}"'.format(include_path))

        # Skip already processed files.
        if file_path in self.__processed_files:
            return
        self.__processed_files.add(file_path)

        for token in tokens:
            # Find local includes.
            token = self.__scan_for_include(file_path, token, is_source_file)
            if not token or self.__scan_for_source_files_stitch(token):
                continue

            self.__write_token(token, is_main_header)

        file_path = self.__find_possible_source_file(file_path)
        if file_path:
            self.__source_files.put(file_path)

    def __write_token(self, token: Token, is_main_header: bool):
        if isinstance(token, StartToken) or isinstance(token, EndToken):
            return

        if (not is_main_header and self.__is_pragma_once(token)) or self.__is_include_guard(token):
            token = token.preprocessor_tokens[-2]
            if not isinstance(token, LinebreakWhitespaceToken):
                return

        if self.__is_cont_line_break(token):
            return

        # Write previous token, store current.
        if self.__prev_token:
            self.__dst.write(str(self.__prev_token.raw))
        self.__prev_token = token

    @staticmethod
    def __is_pragma_once(token: Token):
        if isinstance(token, PreprocessorPragmaOnceToken):
            return True
        return False

    def __is_include_guard(self, token: Token):
        if self.__include_guard_format is None:
            return False

        if isinstance(token, (PreprocessorIfNotDefinedToken, PreprocessorDefineToken)):
            # Find first remaining token matching the include guard format.
            i, remaining_token = find_token(token.preprocessor_arguments, RemainingToken)
            if remaining_token and self.__include_guard_format.match(str(remaining_token).strip()) and \
                    contains_only_whitespace_and_comment_tokens(token.preprocessor_arguments[i + 1:]):
                return True
        elif isinstance(token, PreprocessorEndIfToken):
            # Find first comment token matching the include guard format.
            i, comment_token = find_token(token.preprocessor_arguments, CommentToken)
            if comment_token and self.__include_guard_format.match(str(comment_token.content).strip()) and \
                    contains_only_whitespace_and_comment_tokens(token.preprocessor_arguments[i + 1:]):
                return True

    def __find_possible_source_file(self, header_file_path: Path) -> Union[Path, None]:
        if header_file_path.suffix in ['.c', '.cpp', '.cxx', '.cc', '.c++', '.cp', '.C']:
            return

        # Checks if a equivalent compilation unit exits.
        for extension in ['.c', '.cpp', '.cxx', '.cc', '.c++', '.cp', '.C']:
            for src_dir in self.__relative_source_directories:
                file_path = (header_file_path.parent / src_dir / header_file_path.name).with_suffix(extension)
                if file_path.exists():
                    return file_path
            for src_dir in self.__source_directories:
                file_path = (src_dir / header_file_path.name).with_suffix(extension).resolve()
                if file_path.exists():
                    return file_path
        return None

    def __scan_for_include(self, file_path: Path, token: Token, is_source_file: bool) -> Union[Token, None]:
        if not isinstance(token, PreprocessorIncludeToken) or not token.is_local_include:
            return token

        self.__process_file(file_path.parent, Path(str(token.path)), is_source_file)
        # Take include tokens line break token if any.
        token = token.preprocessor_tokens[-2]
        if isinstance(token, LinebreakWhitespaceToken):
            return token

        return None

    def __scan_for_source_files_stitch(self, token: Token) -> bool:
        if not isinstance(token, CommentToken) or str(token.content).strip() != self.__stitch_format:
            return False

        while not self.__source_files.empty():
            self.__process_file(Path(), self.__source_files.get(), True)

        return True

    def __is_cont_line_break(self, token: Token) -> bool:
        if not self.__trim:
            return False

        if isinstance(token, LinebreakWhitespaceToken):
            self.__cont_lb += 1
        elif isinstance(token, PreprocessorToken) and isinstance(token.preprocessor_tokens[-2],
                                                                 LinebreakWhitespaceToken):
            self.__cont_lb = CONTINUOUS_LINE_BREAK_START + 1
        else:
            self.__cont_lb = CONTINUOUS_LINE_BREAK_START

        return self.__cont_lb >= CONTINUOUS_BREAK_REACHED



from typing import List

from .token import Token
from .iterator import LineWrapIterator


class NumberToken(Token):
    pass


def scan_for_number(tokens: List[Token], it: LineWrapIterator):
    if not it.curr.isdigit() and (it.curr != '.' or not it.lookahead.isdigit()):
        return False
    start = it.copy()

    # Parse until not
    # * alphanumeric, _, .
    # * ' followed by another '
    # * ' followed by an alphanumeric and _
    while it.next() and (it.curr.isalnum() or it.curr in '_+-.' or (
            it.curr == "'" and (it.lookahead == "'" or it.lookahead.isalnum() or it.lookahead in '_.'))):
        pass

    tokens.append(NumberToken(start, it))
    return True



"""
    Setup file for compile-time-printer.
    Use setup.cfg to configure your project.
    This file was generated with PyScaffold 4.0b1.
    PyScaffold helps you to put up the scaffold of your new Python project.
    Learn more under: https://pyscaffold.org/
"""
from setuptools import setup

if __name__ == '__main__':
    try:
        setup(use_scm_version={'version_scheme': 'post-release'})
    except:  # noqa
        print(
            '\n\nAn error occurred while building the project, '
            'please ensure you have the most updated version of setuptools, '
            'setuptools_scm and wheel with:\n'
            '   pip install -U setuptools setuptools_scm wheel\n\n'
        )
        raise



from typing import List

from .iterator import LineWrapIterator
from .token import Token


class WhitespaceToken(Token):
    pass


class WhitespaceWhitespaceToken(WhitespaceToken):
    pass


class LinebreakWhitespaceToken(WhitespaceToken):
    pass


WHITESPACE_CHARACTERS = ' \t\v\f'


def scan_for_whitespace(tokens: List[Token], it: LineWrapIterator):
    if it.curr in WHITESPACE_CHARACTERS:
        start = it.copy()
        while it.next() and it.curr in WHITESPACE_CHARACTERS:
            pass

        tokens.append(WhitespaceWhitespaceToken(start, it))
        return True
    elif it.curr in '\n\r':
        start = it.copy()
        it.next()

        if it.prev == '\r' and it.curr == '\n':
            it.next()

        tokens.append(LinebreakWhitespaceToken(start, it))
        return True
    return False
